/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <log4cxx/logstring.h>
#include <log4cxx/helpers/outputstreamwriter.h>
#include <log4cxx/helpers/exception.h>
#include <log4cxx/helpers/charsetencoder.h>
#include <log4cxx/helpers/bytebuffer.h>
#include <log4cxx/helpers/stringhelper.h>

using namespace LOG4CXX_NS;
using namespace LOG4CXX_NS::helpers;

IMPLEMENT_LOG4CXX_OBJECT(OutputStreamWriter)

struct OutputStreamWriter::OutputStreamWriterPrivate
{
	OutputStreamWriterPrivate
		( const OutputStreamPtr& out1
		, const CharsetEncoderPtr& enc1 = CharsetEncoder::getDefaultEncoder()
		)
		: out(out1)
		, enc(enc1)
		{}

	OutputStreamPtr out;
	CharsetEncoderPtr enc;
};

OutputStreamWriter::OutputStreamWriter(LOG4CXX_16_CONST OutputStreamPtr& out)
	: m_priv(std::make_unique<OutputStreamWriterPrivate>(out))
{
	if (!out)
	{
		throw NullPointerException(LOG4CXX_STR("OutputStream parameter"));
	}
}

OutputStreamWriter::OutputStreamWriter
	( LOG4CXX_16_CONST OutputStreamPtr& out
	, LOG4CXX_16_CONST CharsetEncoderPtr& enc
	)
	: m_priv(std::make_unique<OutputStreamWriterPrivate>(out, enc))
{
	if (!out)
	{
		throw NullPointerException(LOG4CXX_STR("OutputStream parameter"));
	}

	if (!enc)
	{
		throw NullPointerException(LOG4CXX_STR("CharsetEncoder parameter"));
	}
}

OutputStreamWriter::~OutputStreamWriter()
{
}

void OutputStreamWriter::close( LOG4CXX_CLOSE_WRITER_FORMAL_PARAMETERS )
{
	m_priv->out->close();
}

void OutputStreamWriter::flush( LOG4CXX_FLUSH_WRITER_FORMAL_PARAMETERS )
{
	m_priv->out->flush();
}

void OutputStreamWriter::write( LOG4CXX_WRITE_WRITER_FORMAL_PARAMETERS )
{
	if (str.empty())
		return;
	if (CharsetEncoder::isTriviallyCopyable(str, m_priv->enc))
	{
		ByteBuffer buf((char*)str.data(), str.size() * sizeof (logchar));
		m_priv->out->write(buf);
	}
	else
	{
		enum { BUFSIZE = 1024 };
		char stackData[BUFSIZE];
		char* rawbuf = stackData;
		size_t bufSize = BUFSIZE;
#ifdef LOG4CXX_MULTI_PROCESS
		std::vector<char> heapData;
		// Ensure the logging event is a single write system call to keep events from each process separate
		if (bufSize < str.length() * 2)
		{
			heapData.resize(bufSize = str.length() * 2);
			rawbuf = heapData.data();
		}
#endif
		ByteBuffer buf(rawbuf, bufSize);
		m_priv->enc->reset();
		LogString::const_iterator iter = str.begin();

		while (iter != str.end())
		{
			CharsetEncoder::encode(m_priv->enc, str, iter, buf);
			buf.flip();
			m_priv->out->write(buf);
			buf.clear();
		}

		CharsetEncoder::encode(m_priv->enc, str, iter, buf);
		m_priv->enc->flush(buf);
		buf.flip();
		m_priv->out->write(buf);
	}
}

OutputStreamPtr OutputStreamWriter::getOutputStreamPtr() const
{
	return m_priv->out;
}

